import configparser

from mcmd import history, io
from mcmd.utils import McmdError


def execute(args, exit_on_error=True):
    try:
        args.func(args)
    except McmdError as e:
        _handle_error(str(e), args.write_to_history, args.arg_string, exit_on_error)
    except configparser.Error as e:
        message = 'Error reading or writing mcmd.properties: %s' % str(e)
        _handle_error(message, args.write_to_history, args.arg_string, exit_on_error)
    else:
        if args.write_to_history:
            history.write(args.arg_string, success=True)
        io.succeed()


def _handle_error(message, write_to_history, arg_string, exit_on_error):
    io.error(message)
    if write_to_history:
        history.write(arg_string, success=False)
    if exit_on_error:
        exit(1)



import argparse

from mcmd.commands.add import arguments as add_args
from mcmd.commands.disable import arguments as disable_args
from mcmd.commands.enable import arguments as enable_args
from mcmd.commands.give import arguments as give_args
from mcmd.commands.history import arguments as history_args
from mcmd.commands.import_ import arguments as import_args
from mcmd.commands.make import arguments as make_args
from mcmd.commands.run import arguments as run_args
from mcmd.commands.script import arguments as script_args

_parser = None


def _get_parser():
    global _parser
    if not _parser:
        _parser = _create_parser()

    return _parser


def _create_parser():
    parser = argparse.ArgumentParser(prog='mcmd')
    subparsers = parser.add_subparsers(title="commands", dest="command")

    # add general optionals
    parser.add_argument('--as-user', '-u',
                        type=str,
                        metavar='USER',
                        help="Execute a command as a user. (The default user is set in the mcmd.properties file). "
                             "Assumes that the password is the same as the username. If it isn't, also supply the "
                             "--with-password argument.")
    parser.add_argument('--with-password', '-p',
                        type=str,
                        metavar='PASSWORD',
                        help="The password to use when logging in. (The default is set in the mcmd.properties file)")
    parser.add_argument('--verbose', '-v',
                        action='count',
                        help='Print verbose messages')

    # add sub commands
    import_args(subparsers)
    make_args(subparsers)
    add_args(subparsers)
    give_args(subparsers)
    enable_args(subparsers)
    disable_args(subparsers)
    run_args(subparsers)
    script_args(subparsers)
    history_args(subparsers)

    return parser


def parse_args():
    return _get_parser().parse_args()


def parse_arg_string(argument_string):
    return _get_parser().parse_args(argument_string)


def print_help():
    _get_parser().print_help()
    exit(1)



from pathlib import Path

from mcmd import arguments as arg_parser
from mcmd import io
from mcmd.executor import execute


# =========
# Arguments
# =========

def arguments(subparsers):
    p_run = subparsers.add_parser('run',
                                  help='Run an mcmd script')
    p_run.set_defaults(write_to_history=False)
    p_run.add_argument('script',
                       type=str,
                       help='The .mcmd script to run')
    p_run.add_argument('--ignore-errors', '-i',
                       action='store_true',
                       help='Let the script continue when one or more commands throw an error')


# =======
# Methods
# =======

def run(args):
    script = Path().home().joinpath('.mcmd', 'scripts', args.script)

    lines = list()
    try:
        with open(script) as file:
            lines = [line.rstrip('\n') for line in file]
    except OSError as e:
        io.error('Error reading script: %s' % str(e))

    exit_on_error = not args.ignore_errors
    for line in lines:
        sub_args = arg_parser.parse_arg_string(line.split(' '))
        setattr(sub_args, 'arg_string', line)
        if sub_args.command == 'run':
            io.error("Can't use the run command in a script: %s" % line)
            if exit_on_error:
                exit(1)
            else:
                continue

        execute(sub_args, exit_on_error)



from mcmd import io
from mcmd.client.molgenis_client import login, ResourceType, post, ensure_resource_exists
from mcmd.config.config import config
from mcmd.io import highlight


# =========
# Arguments
# =========


def arguments(subparsers):
    p_add = subparsers.add_parser('disable',
                                  help='Disable resources/functionality',
                                  description="Run 'mcmd disable rls -h' to view the help for those sub-commands")
    p_add_subparsers = p_add.add_subparsers(dest="type")

    p_add_group = p_add_subparsers.add_parser('rls',
                                              help='Disables row level security on an entity type')
    p_add_group.set_defaults(func=enable_rls,
                             write_to_history=True)
    p_add_group.add_argument('entity',
                             type=str,
                             help="The entity type to remove the row level security from")


# =======
# Methods
# =======

@login
def enable_rls(args):
    if not io.confirm('Are you sure you want to disable row level security on %s?' % args.entity):
        return

    io.start('Disabling row level security on entity type %s' % highlight(args.entity))

    ensure_resource_exists(args.entity, ResourceType.ENTITY_TYPE)
    post(config().get('api', 'rls'), data={'id': args.entity,
                                           'rlsEnabled': False})



import shutil
from configparser import ConfigParser

import pkg_resources

from mcmd import io
from mcmd.config.home import get_properties_file
from mcmd.utils import McmdError

_DEFAULT_PROPERTIES = pkg_resources.resource_stream('mcmd.config', 'default.properties')

_config = None


def _create_user_config():
    try:
        with get_properties_file().open('wb') as properties_file:
            shutil.copyfileobj(_DEFAULT_PROPERTIES, properties_file)
    except OSError as err:
        raise McmdError("Error creating properties file: %s" % err)


def _check_user_config():
    if not get_properties_file().exists():
        io.info("No properties file found. Creating it now.")
        _create_user_config()


def _load_config():
    global _config
    _check_user_config()
    _config = ConfigParser()
    _config.read(get_properties_file())


def config():
    if not _config:
        _load_config()
    return _config



import re

from github import Github, UnknownObjectException

from mcmd.utils import McmdError

_MOLGENIS_FILES_URL = 'https://github.com/molgenis/molgenis/files/'

_github = None


class Attachment:
    def __init__(self, url):
        self.url = url
        self.id = self.parse_id(url)
        self.name = self.parse_file_name(url)

    def __repr__(self):
        return 'Attachment: %s' % self.id

    @staticmethod
    def parse_id(url):
        """Parses the file name, extension and the preceding identifier from the url."""
        return '/'.join(url.rsplit('/', 2)[-2:])

    @staticmethod
    def parse_file_name(url):
        """Parses the file name + extension from the url."""
        return '/'.join(url.rsplit('/', 1)[-1:])


def get_attachments(issue_num):
    validate_issue_number(issue_num)
    try:
        issue = _molgenis_repo().get_issue(int(issue_num))
    except UnknownObjectException:
        raise McmdError("Issue #%s doesn't exist" % issue_num)

    # GitHub has no API for downloading attachments so we get them from the issue description
    urls = _parse_attachment_urls(issue.body)
    return [Attachment(url.strip('()')) for url in urls]


def _parse_attachment_urls(issue_body):
    return re.findall('\((%s.*?)\)' % re.escape(_MOLGENIS_FILES_URL), issue_body)


def _molgenis_repo():
    global _github
    if not _github:
        _github = Github()

    return _github.get_organization('molgenis').get_repo('molgenis')


def validate_issue_number(issue_num):
    try:
        int(issue_num)
    except ValueError:
        raise McmdError('Not a valid issue number: %s' % issue_num)



from mcmd import history as hist
from mcmd import io
from mcmd.logging import get_logger


# =========
# Arguments
# =========

def arguments(subparsers):
    p_history = subparsers.add_parser('history',
                                      help="Shows the history of commands that were run. Commands from scripts are"
                                           " also included.")
    p_history.set_defaults(func=history,
                           write_to_history=False)
    p_history.add_argument('--number', '-n',
                           type=int,
                           default=10,
                           help='Number of lines of history to show. Default: 10')
    p_history.add_argument('--clear', '-c',
                           action='store_true',
                           help='Clears the history.')


# =======
# Globals
# =======

log = get_logger()


# =======
# Methods
# =======

def history(args):
    if args.clear:
        io.start('Clearing history')
        hist.clear()
    else:
        lines = hist.read(args.number, include_fails=True)
        if len(lines) == 0:
            log.info('History is empty.')
        for line in lines:
            io.start(line[1])
            if line[0]:
                io.succeed()
            else:
                io.error(None)



import json
from enum import Enum

import requests

from mcmd import io
from mcmd.config.config import config
from mcmd.logging import get_logger
from mcmd.utils import McmdError

log = get_logger()
token = ''


class ResourceType(Enum):
    ENTITY_TYPE = ('sys_md_EntityType', 'entityclass', 'Entity Type')
    PACKAGE = ('sys_md_Package', 'package', 'Package')
    PLUGIN = ('sys_Plugin', 'plugin', 'Plugin')

    def get_entity_id(self):
        return self.value[0]

    def get_resource_name(self):
        return self.value[1]

    def get_label(self):
        return self.value[2]

    @classmethod
    def of_label(cls, label):
        return ResourceType[label.replace(' ', '_').upper()]


class PrincipalType(Enum):
    USER = 'user'
    ROLE = 'role'


def login(func):
    """Login decorator."""

    def wrapper(args):
        global token
        if args.as_user is None:
            username = config().get('auth', 'username')
        else:
            username = args.as_user

        if args.with_password is None:
            if args.as_user is None:
                password = config().get('auth', 'password')
            else:
                password = args.as_user
        else:
            password = args.with_password

        login_url = config().get('api', 'login')

        io.debug('Logging in as user %s' % username)

        response = post(login_url,
                        data={"username": username, "password": password})
        token = response.json()['token']

        func(args)

    return wrapper


def get(url):
    return _handle_request(lambda: requests.get(url,
                                                headers=_get_default_headers()))


def grant(principal_type, principal_name, resource_type, identifier, permission):
    data = {'radio-' + identifier: permission}

    if principal_type == PrincipalType.USER:
        data['username'] = principal_name
    elif principal_type == PrincipalType.ROLE:
        data['rolename'] = principal_name.upper()
    else:
        raise McmdError('Unknown principal type: %s' % principal_type)

    url = config().get('api', 'perm') + resource_type.get_resource_name() + '/' + principal_type.value
    return _handle_request(lambda: requests.post(url,
                                                 headers={
                                                     'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                                                     'x-molgenis-token': token},
                                                 data=data))


def post(url, data):
    return _handle_request(lambda: requests.post(url,
                                                 headers=_get_default_headers(),
                                                 data=json.dumps(data)))


def post_file(url, file_path, params):
    return _handle_request(lambda: requests.post(url,
                                                 headers={'x-molgenis-token': token},
                                                 files={'file': open(file_path, 'rb')},
                                                 params=params))


def import_by_url(params):
    return _handle_request(lambda: requests.post(config().get('api', 'import_url'),
                                                 headers=_get_default_headers(),
                                                 params=params))


def _get_default_headers():
    headers = {'Content-Type': 'application/json'}
    if token:
        headers['x-molgenis-token'] = token
    return headers


def _handle_request(request):
    response = str()
    try:
        response = request()
        response.raise_for_status()
        return response
    except requests.HTTPError as e:
        if _is_json(response):
            _handle_json_error(response.json())
        raise McmdError(str(e))
    except requests.RequestException as e:
        raise McmdError(str(e))


def _handle_json_error(response_json):
    if 'errors' in response_json:
        for error in response_json['errors']:
            # TODO capture multiple error messages
            raise McmdError(error['message'])
    elif 'errorMessage' in response_json:
        raise McmdError(response_json['errorMessage'])


def _is_json(response):
    return response.headers.get('Content-Type') and 'application/json' in response.headers.get('Content-Type')


def resource_exists(resource_id, resource_type):
    log.debug('Checking if %s %s exists' % (resource_type.get_label(), resource_id))
    response = get(config().get('api', 'rest2') + resource_type.get_entity_id() + '?q=id==' + resource_id)
    return int(response.json()['total']) > 0


def ensure_resource_exists(resource_id, resource_type):
    if not resource_exists(resource_id, resource_type):
        raise McmdError('No %s found with id %s' % (resource_type.get_label(), resource_id))


def ensure_principal_exists(principal_name, principal_type):
    if not principal_exists(principal_name, principal_type):
        raise McmdError('No {} found with id {}'.format(principal_type, principal_name))


def principal_exists(principal_name, principal_type):
    if principal_type == PrincipalType.USER:
        return user_exists(principal_name)
    elif principal_type == PrincipalType.ROLE:
        return role_exists(principal_name)
    else:
        raise McmdError('Unknown principal type: %s' % principal_type)


def user_exists(username):
    log.debug('Checking if user %s exists' % username)
    response = get(config().get('api', 'rest2') + 'sys_sec_User?q=username==' + username)
    return int(response.json()['total']) > 0


def role_exists(rolename):
    log.debug('Checking if role %s exists' % rolename)
    response = get(config().get('api', 'rest2') + 'sys_sec_Role?q=name==' + rolename.upper())
    return int(response.json()['total']) > 0



from collections import defaultdict
from pathlib import Path
from urllib.parse import urljoin

import polling
import requests

from mcmd import io
from mcmd.client import github_client as github
from mcmd.client.molgenis_client import login, post_file, get, import_by_url
from mcmd.config.config import config
from mcmd.config.home import get_issues_folder
from mcmd.io import highlight
from mcmd.utils import McmdError, config_string_to_paths


# =========
# Arguments
# =========

def arguments(subparsers):
    p_import = subparsers.add_parser('import',
                                     help='Import a file')
    p_import.set_defaults(func=import_,
                          write_to_history=True)
    p_import.add_argument('file',
                          nargs='?',
                          help='The file to upload')
    p_import_source = p_import.add_mutually_exclusive_group()
    p_import_source.add_argument('--from-path', '-p',
                                 action='store_true',
                                 help='Import a file the old school way (by path)')
    p_import_source.add_argument('--from-issue', '-i',
                                 metavar='ISSUE_NUMBER',
                                 help='Import a file from a GitHub issue')
    p_import_source.add_argument('--from-url',
                                 metavar='URL',
                                 help='Import a file from a URL. Uses the importByUrl endpoint of the MOLGENIS '
                                      'importer.')
    p_import.add_argument('--in',
                          dest='to_package',
                          type=str,
                          metavar='PACKAGE_ID',
                          help='The package to import to')
    return p_import


# =======
# Methods
# =======

@login
def import_(args):
    if args.from_path:
        _import_from_path(args)
    elif args.from_url:
        _import_from_url(args)
    elif args.from_issue:
        _import_from_issue(args)
    else:
        _import_from_quick_folders(args)


def _import_from_url(args):
    file_url = args.from_url
    file_name = file_url.split("/")[-1]
    io.start('Importing from URL %s' % highlight(file_url))

    params = {'action': _get_import_action(file_name),
              'metadataAction': 'upsert'}

    if args.to_package:
        params['packageId'] = args.to_package

    params['url'] = file_url

    response = import_by_url(params)
    import_run_url = urljoin(config().get('api', 'host'), response.text)
    status, message = _poll_for_completion(import_run_url)
    if status == 'FAILED':
        raise McmdError(message)


def _import_from_quick_folders(args):
    file_name = args.file
    file_map = _scan_folders_for_files(_get_molgenis_folders() + _get_quick_folders())
    path = _select_path(file_map, file_name)
    _do_import(path, args.to_package)


def _import_from_issue(args):
    issue_num = args.from_issue
    attachment = _select_attachment(issue_num, args.file)
    file_path = _download_attachment(attachment, issue_num)
    _do_import(file_path, args.to_package)


def _import_from_path(args):
    io.start('Importing from path %s' % highlight(args.file))
    file = Path(args.file)
    if not file.is_file():
        raise McmdError("File %s doesn't exist" % str(file.resolve()))
    _do_import(file, args.to_package)


def _select_path(file_map, file_name):
    if file_name in file_map:
        paths = file_map[file_name]
        if len(paths) > 1:
            path = _choose_file(paths, file_name)
        else:
            path = paths[0]
    else:
        raise McmdError('No file found for %s' % file_name)
    return path


def _select_attachment(issue_num, wanted_attachment):
    """Gets attachments from a GitHub issue. If wanted_attachment is specified it will try to select that attachment."""
    attachments = github.get_attachments(issue_num)
    if len(attachments) == 0:
        raise McmdError("Issue #%s doesn't contain any files" % issue_num)

    if wanted_attachment:
        selected = [a for a in attachments if a.name == wanted_attachment]
        if len(selected) == 0:
            raise McmdError('There are no attachments named %s.' % wanted_attachment)
        if len(selected) > 1:
            raise McmdError('Multiple attachments with name %s found.' % wanted_attachment)
        return selected[0]
    else:
        if len(attachments) > 1:
            return _choose_attachment(attachments)
        else:
            return attachments[0]


def _create_attachment_map(attachments):
    """"Creates a dict of name/attachment pairs. If the name isn't unique, uses the identifier as key. Example:
            1234/example.xls
            4567/example.xls
            other_example.xls
    """
    names = [a.name for a in attachments]

    attachment_map = dict()
    for a in attachments:
        if names.count(a.name) > 1:
            attachment_map[a.id] = a
        else:
            attachment_map[a.name] = a

    return attachment_map


def _choose_attachment(attachments):
    """Let user choose from multiple attachments. Attachments with duplicate names will display their id."""
    attachment_map = _create_attachment_map(attachments)
    choices = list(attachment_map.keys())

    answer = io.multi_choice('Multiple attachments found. Choose which one to import:', choices)
    return attachment_map[answer]


def _choose_file(paths, name):
    choices = [str(path) for path in paths]
    answer = io.multi_choice('Multiple files found for %s. Pick one:' % name, choices)
    return Path(answer)


def _download_attachment(attachment, issue_num):
    issue_folder = get_issues_folder().joinpath(issue_num)
    issue_folder.mkdir(parents=True, exist_ok=True)
    file_path = issue_folder.joinpath(attachment.name)

    if file_path.exists():
        overwrite = io.confirm('File %s already exists. Re-download?' % file_path.name)
        if not overwrite:
            return file_path

    io.start('Downloading %s from GitHub issue %s' % (highlight(attachment.name), highlight('#' + issue_num)))
    try:
        r = requests.get(attachment.url)
        r.raise_for_status()
        with file_path.open('wb') as f:
            f.write(r.content)
    except (OSError, requests.RequestException, requests.HTTPError) as e:
        raise McmdError('Error downloading GitHub attachment: %s' % str(e))
    io.succeed()
    return file_path


def _do_import(file_path, package):
    io.start('Importing %s' % (highlight(file_path.name)))

    params = {'action': _get_import_action(file_path.name),
              'metadataAction': 'upsert'}

    if package:
        params['packageId'] = package

    response = post_file(config().get('api', 'import'), file_path.resolve(), params)
    import_run_url = urljoin(config().get('api', 'host'), response.text)
    status, message = _poll_for_completion(import_run_url)
    if status == 'FAILED':
        raise McmdError(message)


def _get_import_action(file_name):
    if '.owl' in file_name or '.obo' in file_name:
        return 'add'
    else:
        return config().get('set', 'import_action')


def _poll_for_completion(url):
    polling.poll(lambda: get(url).json()['status'] != 'RUNNING',
                 step=0.1,
                 poll_forever=True)
    import_run = get(url).json()
    return import_run['status'], import_run['message']


def _scan_folders_for_files(folders):
    files = defaultdict(list)
    for folder in folders:
        if not folder.is_dir():
            io.warn('Folder %s is not a valid folder, skipping it...' % folder)

        for file in list(folder.glob('*.*')):
            files[file.stem].append(file)
    return files


def _get_molgenis_folders():
    if not config().has_option('data', 'git_root') or not config().has_option('data', 'git_paths'):
        io.info('Molgenis git paths not configured. Edit the mcmd.properties file to include the test data.')
        return list()
    else:
        return config_string_to_paths(config().get('data', 'git_paths'))


def _get_quick_folders():
    if not config().has_option('data', 'quick_folders'):
        return list()
    else:
        return config_string_to_paths(config().get('data', 'quick_folders'))



from mcmd import io
from mcmd.client.molgenis_client import login, post, get
from mcmd.config.config import config
from mcmd.io import highlight
from mcmd.utils import McmdError


# =========
# Arguments
# =========

def arguments(subparsers):
    p_add = subparsers.add_parser('add',
                                  help='Add users and groups',
                                  description="Run 'mcmd add group -h' or 'mcmd add user -h' to view the help for those"
                                              " sub-commands")
    p_add_subparsers = p_add.add_subparsers(dest="type")

    p_add_group = p_add_subparsers.add_parser('group',
                                              help='Add a group')
    p_add_group.set_defaults(func=add_group,
                             write_to_history=True)
    p_add_group.add_argument('name',
                             type=str,
                             help="The group's name")

    p_add_user = p_add_subparsers.add_parser('user',
                                             help='Add a user')
    p_add_user.set_defaults(func=add_user,
                            write_to_history=True)
    p_add_user.add_argument('username',
                            type=str,
                            help="The user's name")
    p_add_user.add_argument('--set-password', '-p',
                            metavar='PASSWORD',
                            type=str,
                            help="The user's password")
    p_add_user.add_argument('--with-email', '-e',
                            metavar='EMAIL',
                            type=str,
                            help="The user's e-mail address")
    p_add_user.add_argument('--is-inactive', '-a',
                            action='store_true',
                            help="Make user inactive")
    p_add_user.add_argument('--is-superuser', '-s',
                            action='store_true',
                            help="Make user superuser")
    p_add_user.add_argument('--change-password', '-c',
                            action='store_true',
                            help="Set change password to true for user")

    p_add_package = p_add_subparsers.add_parser('package',
                                                help='Add a package')
    p_add_package.set_defaults(func=add_package,
                               write_to_history=True)
    p_add_package.add_argument('id',
                               type=str,
                               help="The id of the Package")
    p_add_package.add_argument('--in',
                               type=str,
                               dest='parent',
                               help="The id of the parent")

    p_add_token = p_add_subparsers.add_parser('token',
                                              help='Add a token')
    p_add_token.set_defaults(func=add_token,
                             write_to_history=True)
    p_add_token.add_argument('user',
                             type=str,
                             help="The user to give the token to")
    p_add_token.add_argument('token',
                             type=str,
                             help="The token")


# =======
# Methods
# =======

@login
def add_user(args):
    io.start('Adding user %s' % highlight(args.username))

    password = args.set_password if args.set_password else args.username
    email = args.with_email if args.with_email else args.username + '@molgenis.org'
    active = not args.is_inactive
    superuser = args.is_superuser
    ch_pwd = args.change_password

    post(config().get('api', 'rest1') + 'sys_sec_User',
         {'username': args.username,
          'password_': password,
          'changePassword': ch_pwd,
          'Email': email,
          'active': active,
          'superuser': superuser
          })


@login
def add_group(args):
    io.start('Adding group %s' % highlight(args.name))
    post(config().get('api', 'group'), {'name': args.name.lower(), 'label': args.name})


@login
def add_package(args):
    io.start('Adding package %s' % highlight(args.id))

    data = {'id': args.id,
            'label': args.id}

    if args.parent:
        data['parent'] = args.parent

    post(config().get('api', 'rest1') + 'sys_md_Package', data)


@login
def add_token(args):
    io.start('Adding token %s for user %s' % (highlight(args.token), highlight(args.user)))

    user = get(config().get('api', 'rest2') + 'sys_sec_User?attrs=id&q=username==%s' % args.user)
    if user.json()['total'] == 0:
        raise McmdError('Unknown user %s' % args.user)

    user_id = user.json()['items'][0]['id']

    data = {'User': user_id,
            'token': args.token}

    post(config().get('api', 'rest1') + 'sys_sec_Token', data)



from setuptools import setup

setup(
    name='molgenis-commander',
    version='0.1.0',
    packages=['mcmd', 'mcmd.client', 'mcmd.commands', 'mcmd.config'],
    description='A command line interface for Molgenis',
    url='https://github.com/molgenis/molgenis-tools-commander',
    author='Tommy de Boer',
    author_email='tommydeboer4@gmail.com',
    license='GNU Lesser General Public License 3.0',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mcmd = mcmd.__main__:main'
        ]
    },
    install_requires=['requests==2.21.0', 'rainbow_logging_handler==2.2.2', 'PyInquirer==1.0.3', 'halo==0.0.22',
                      'polling==0.3.0', 'PyGithub==1.43.3', 'colorama==0.3.9']
)



import logging
import signal
import sys

from mcmd import io
from mcmd.arguments import parse_args, print_help
from mcmd.commands.run import run
from mcmd.executor import execute
from mcmd.io import set_debug
from mcmd.logging import set_level


def main():
    # setup friendly interrupt message
    signal.signal(signal.SIGINT, interrupt_handler)

    args = parse_args()
    if not args.command:
        print_help()

    setattr(args, 'arg_string', ' '.join(sys.argv[1:]))
    set_log_level(args)

    if args.command == 'run':
        run(args)
    else:
        execute(args)


def set_log_level(args):
    if args.verbose:
        verbosity = int(args.verbose)
        if verbosity > 0:
            set_level(logging.DEBUG)
            set_debug()
    else:
        set_level(logging.INFO)


# noinspection PyUnusedLocal
def interrupt_handler(sig, frame):
    io.warn('Interrupted by user.')
    sys.exit(0)


if __name__ == '__main__':
    main()



from pathlib import Path


class McmdError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return repr(self.message)


def upper_snake(string):
    """
    Transforms a string to uppercase snake style.
    E.g. 'lower-kebab-style' becomes 'LOWER_KEBAB_STYLE'.
    """
    return string.upper().replace('-', '_')


def lower_kebab(string):
    """
    Transforms a string to lowercase kebab style.
    E.g. 'SCREAMING_SNAKE' becomes 'screaming-snake'.
    """
    return string.lower().replace('_', '-')


def config_string_to_paths(config_string):
    """Strips and splits comma separated, multi-line configuration variables and converts them to Path objects."""
    clean_string = ''.join(config_string.split())
    paths = clean_string.split(',')
    return [Path(path_string) for path_string in paths]



"""
Give a principal (a user or a group) some permission on a resource (a package, entity type or plugin). Unless specified
by the user, the give command will try to figure out the principal- and resource types itself. If a resource or
principal doesn't exist, the program will terminate.
"""

from mcmd import io
from mcmd.client.molgenis_client import login, grant, PrincipalType, principal_exists, \
    resource_exists, ResourceType, ensure_resource_exists, ensure_principal_exists
from mcmd.io import multi_choice, highlight
from mcmd.utils import McmdError


# =========
# Arguments
# =========

def arguments(subparsers):
    p_give = subparsers.add_parser('give',
                                   help='Give permissions on resources to roles or users.')
    p_give.set_defaults(func=give,
                        write_to_history=True)
    p_give_resource = p_give.add_mutually_exclusive_group()
    p_give_resource.add_argument('--entity-type', '-e',
                                 action='store_true',
                                 help='Flag to specify that the resource is an entity type')
    p_give_resource.add_argument('--package', '-p',
                                 action='store_true',
                                 help='Flag to specify that the resource is a package')
    p_give_resource.add_argument('--plugin', '-pl',
                                 action='store_true',
                                 help='Flag to specify that the resource is a plugin')
    p_give_receiver = p_give.add_mutually_exclusive_group()
    p_give_receiver.add_argument('--user', '-u',
                                 action='store_true',
                                 help='Flag to specify that the receiver is a user')
    p_give_receiver.add_argument('--role', '-r',
                                 action='store_true',
                                 help='Flag to specify that the receiver is a role')
    p_give.add_argument('receiver',
                        type=str,
                        help='The role (or user) to give the permission to')
    p_give.add_argument('permission',
                        choices=['none', 'writemeta', 'readmeta', 'write', 'edit', 'read', 'view', 'count'],
                        help='The permission type to give. Synonyms are allowed (e.g. write/edit).')
    p_give.add_argument('resource',
                        type=str,
                        help='The resource to which permission is given')


# =======
# Globals
# =======


_PERMISSION_SYNONYMS = {'view': 'read',
                        'edit': 'write'}


# =======
# Methods
# =======

@login
def give(args):
    # Convert synonyms to correct permission type
    if args.permission in _PERMISSION_SYNONYMS:
        args.permission = _PERMISSION_SYNONYMS[args.permission]

    # The PermissionManagerController always gives 200 OK so we need to validate everything ourselves
    resource_type = _get_resource_type(args)
    principal_type = _get_principal_type(args)
    io.start('Giving %s %s permission to %s on %s %s' % (principal_type.value,
                                                         highlight(args.receiver),
                                                         highlight(args.permission),
                                                         resource_type.get_label().lower(),
                                                         highlight(args.resource)))

    grant(principal_type, args.receiver, resource_type, args.resource, args.permission)


def _get_principal_type(args):
    principal_name = args.receiver
    if args.user:
        ensure_principal_exists(args.user, principal_name)
        return PrincipalType.USER
    elif args.role:
        ensure_principal_exists(args.role, principal_name)
        return PrincipalType.ROLE
    else:
        return _guess_principal_type(principal_name)


def _guess_principal_type(principal_name):
    results = dict()
    for principal_type in PrincipalType:
        if principal_exists(principal_name, principal_type):
            results[principal_type.value] = principal_name

    if len(results) == 0:
        raise McmdError('No principals found with name %s' % principal_name)
    elif len(results) > 1:
        choices = results.keys()
        answer = multi_choice('Multiple principals found with name %s. Choose one:' % principal_name, choices)
        return PrincipalType[answer.upper()]
    else:
        return PrincipalType[list(results)[0].upper()]


def _get_resource_type(args):
    resource_id = args.resource
    if args.entity_type:
        ensure_resource_exists(resource_id, ResourceType.ENTITY_TYPE)
        return ResourceType.ENTITY_TYPE
    elif args.package:
        ensure_resource_exists(resource_id, ResourceType.PACKAGE)
        return ResourceType.PACKAGE
    elif args.plugin:
        ensure_resource_exists(resource_id, ResourceType.PLUGIN)
        return ResourceType.PLUGIN
    else:
        return _guess_resource_type(resource_id)


def _guess_resource_type(resource_id):
    results = dict()
    for resource_type in ResourceType:
        if resource_exists(resource_id, resource_type):
            results[resource_type.get_label()] = resource_id

    if len(results) == 0:
        raise McmdError('No resources found with id %s' % resource_id)
    elif len(results) > 1:
        choices = results.keys()
        answer = multi_choice('Multiple resources found for id %s. Choose one:' % resource_id, choices)
        return ResourceType.of_label(answer)
    else:
        return ResourceType.of_label(list(results)[0])
